from dataclasses import asdict, fields
from typing import List, Optional

import numpy as np
from scale_sensor_fusion_io.models import (
    Annotation,
    AttributePath,
    AttributesAnnotation,
    Box2DAnnotation,
    Box2DPath,
    CameraDistortion,
    CameraIntrinsics,
    CameraSensor,
    CameraSensorImage,
    CameraSensorVideo,
    CuboidActivationPath,
    CuboidAnnotation,
    CuboidPath,
    CuboidProjectionPath,
    EventAnnotation,
    LabeledPoint,
    LabeledPointsAnnotation,
    LidarSensor,
    LidarSensorFrame,
    LidarSensorPoints,
    LocalizationAdjustmentAnnotation,
    ObjectAnnotation,
    PolygonAnnotation,
    PolygonPath,
    PosePath,
    RadarSensor,
    RadarSensorFrame,
    RadarSensorPoints,
    Scene,
    Sensor,
)
from scale_sensor_fusion_io.spec import SFS


def from_lidarpoints_sfs(points: SFS.LidarSensorPoints) -> LidarSensorPoints:
    return LidarSensorPoints(
        positions=points.positions.reshape(-1, 3),
        colors=points.colors.reshape(-1, 3) if points.colors is not None else None,
        intensities=points.intensities,
        timestamps=points.timestamps if points.timestamps is not None else None,
    )


def to_lidarpoints_sfs(points: LidarSensorPoints) -> SFS.LidarSensorPoints:
    return SFS.LidarSensorPoints(
        positions=points.positions.reshape(-1, 3),
        colors=points.colors.reshape(-1, 3) if points.colors is not None else None,
        intensities=points.intensities,
        timestamps=points.timestamps,
    )


def from_radarpoints_sfs(points: SFS.RadarSensorPoints) -> RadarSensorPoints:
    return RadarSensorPoints(
        positions=points.positions.reshape(-1, 3),
        directions=(
            points.directions.reshape(-1, 3) if points.directions is not None else None
        ),
        timestamps=points.timestamps if points.timestamps is not None else None,
    )


def to_radarpoints_sfs(points: RadarSensorPoints) -> SFS.RadarSensorPoints:
    return SFS.RadarSensorPoints(
        positions=points.positions.reshape(-1, 3),
        directions=(
            points.directions.reshape(-1, 3) if points.directions is not None else None
        ),
        timestamps=points.timestamps,
    )


def from_pose_sfs(posePath: SFS.PosePath) -> PosePath:
    timestamps = np.array(t for t in posePath.timestamps)
    values = np.vstack([v for v in posePath.values])

    return PosePath(data=values, index=timestamps)


def to_pose_sfs(poses: PosePath) -> SFS.PosePath:
    return SFS.PosePath(
        timestamps=poses.index.tolist(),
        values=poses.values.tolist(),
    )


def from_camera_timestamp(camera_sensor: SFS.CameraSensor) -> List[int]:
    if camera_sensor.images:
        return [i.timestamp for i in camera_sensor.images]
    elif camera_sensor.video:
        return camera_sensor.video.timestamps
    else:
        raise Exception("Camera sensor has no timestamp info")


def from_intrinsics_sfs(intrinsics: SFS.CameraIntrinsics) -> CameraIntrinsics:
    return CameraIntrinsics(
        fx=intrinsics.fx,
        fy=intrinsics.fy,
        cx=intrinsics.cx,
        cy=intrinsics.cy,
        width=intrinsics.width,
        height=intrinsics.height,
        distortion=from_distortion_sfs(intrinsics.distortion),
    )


def to_intrinsics_sfs(intrinsics: CameraIntrinsics) -> SFS.CameraIntrinsics:
    return SFS.CameraIntrinsics(
        fx=intrinsics.fx,
        fy=intrinsics.fy,
        cx=intrinsics.cx,
        cy=intrinsics.cy,
        width=intrinsics.width,
        height=intrinsics.height,
        distortion=to_distortion_sfs(intrinsics.distortion),
    )


"""
NOTE: The distortion conversion does not need to keep track of the order of the
coefficients since this is managed by the dataclass themselves. However this implies
that 1) distortion parameters must remain dataclasses and 2) the ordering of the
variables matters.
"""


def from_distortion_sfs(
    distortion: Optional[SFS.CameraDistortion],
) -> Optional[CameraDistortion]:
    if distortion is None or not distortion.model:
        return None

    return CameraDistortion.from_values(
        model=distortion.model, values=distortion.params
    )


def to_distortion_sfs(
    distortion: Optional[CameraDistortion],
) -> Optional[SFS.CameraDistortion]:
    if not distortion or not distortion.model:
        return None

    # dataclasses.asdict may not preserve order so use dataclasses.fields instead
    params = [
        getattr(distortion.params, k.name)
        for k in fields(distortion.params)
        if k.name != "model"
    ]
    return SFS.CameraDistortion(model=distortion.model.value, params=params)


def from_attribute_paths_sfs(
    attributes: Optional[List[SFS.AttributePath]],
) -> Optional[List[AttributePath]]:
    """Convert a list of SFS.AttributePath into SFIO.AttributePath"""

    return (
        [
            AttributePath(
                name=attribute.name,
                timestamps=attribute.timestamps,
                values=attribute.values,
                sensor_id=attribute.sensor_id,
                static=attribute.static,
            )
            for attribute in attributes
        ]
        if attributes
        else None
    )


def to_attribute_paths_sfs(
    attributes: Optional[List[AttributePath]],
) -> Optional[List[SFS.AttributePath]]:
    """Convert a list of SFIO.AttributePath into SFS.AttributePath"""

    return (
        [
            SFS.AttributePath(
                name=attribute.name,
                timestamps=attribute.timestamps,
                values=attribute.values,
                sensor_id=attribute.sensor_id,
                static=attribute.static,
            )
            for attribute in attributes
        ]
        if attributes
        else None
    )


def from_sensors_sfs(sensors: Optional[List[SFS.Sensor]]) -> Optional[List[Sensor]]:
    """Convert a list of SFS.Sensor into SFIO.Sensor"""

    if not sensors:
        return None
    ret: List[Sensor] = []
    for sensor in sensors:
        if isinstance(sensor, SFS.LidarSensor):
            lidar_sensor = sensor
            poses = from_pose_sfs(lidar_sensor.poses)
            frames = [
                LidarSensorFrame(
                    timestamp=int(f.timestamp), points=from_lidarpoints_sfs(f.points)
                )
                for f in lidar_sensor.frames
            ]
            l_sensor = LidarSensor(id=lidar_sensor.id, poses=poses, frames=frames)
            ret.append(l_sensor)
        elif isinstance(sensor, SFS.RadarSensor):
            radar_sensor = sensor
            r_poses = from_pose_sfs(radar_sensor.poses)

            r_frames = [
                RadarSensorFrame(
                    timestamp=int(f.timestamp), points=from_radarpoints_sfs(f.points)
                )
                for f in radar_sensor.frames
            ]
            r_sensor = RadarSensor(id=radar_sensor.id, poses=r_poses, frames=r_frames)
            ret.append(r_sensor)
        elif isinstance(sensor, SFS.CameraSensor):
            camera_sensor = sensor
            c_poses = from_pose_sfs(camera_sensor.poses)
            intrinsics = from_intrinsics_sfs(camera_sensor.intrinsics)
            video, images = None, None

            # sfio and SFS video/images are exactly the same but separated
            # for spec purposes
            if camera_sensor.video:
                video = CameraSensorVideo(**asdict(camera_sensor.video))

            if camera_sensor.images:
                images = [CameraSensorImage(**asdict(i)) for i in camera_sensor.images]

            ret.append(
                CameraSensor(
                    id=camera_sensor.id,
                    poses=c_poses,
                    intrinsics=intrinsics,
                    video=video,
                    images=images,
                )
            )
    return ret


def to_sensors_sfs(sensors: Optional[List[Sensor]]) -> Optional[List[SFS.Sensor]]:
    """Convert a list of SFIO.Sensor into SFS.Sensor"""

    if not sensors:
        return None
    ret: List[SFS.Sensor] = []
    for sensor in sensors:
        if isinstance(sensor, CameraSensor):
            video, images = None, None

            if sensor.video:
                video = SFS.CameraSensorVideo(**asdict(sensor.video))

            if sensor.images:
                images = [SFS.CameraSensorImage(**asdict(i)) for i in sensor.images]

            ret.append(
                SFS.CameraSensor(
                    id=sensor.id,
                    poses=to_pose_sfs(sensor.poses),
                    intrinsics=to_intrinsics_sfs(sensor.intrinsics),
                    video=video,
                    images=images,
                )
            )
        elif isinstance(sensor, RadarSensor):
            ret.append(
                SFS.RadarSensor(
                    id=sensor.id,
                    poses=to_pose_sfs(sensor.poses),
                    frames=[
                        SFS.RadarSensorFrame(
                            timestamp=frame.timestamp,
                            points=to_radarpoints_sfs(frame.points),
                        )
                        for frame in sensor.frames
                    ],
                )
            )
        elif isinstance(sensor, LidarSensor):
            ret.append(
                SFS.LidarSensor(
                    id=sensor.id,
                    poses=to_pose_sfs(sensor.poses),
                    frames=[
                        SFS.LidarSensorFrame(
                            timestamp=frame.timestamp,
                            points=to_lidarpoints_sfs(frame.points),
                        )
                        for frame in sensor.frames
                    ],
                )
            )
    return ret


def from_annotations_sfs(
    annotations: Optional[List[SFS.Annotation]],
) -> Optional[List[Annotation]]:
    """Convert a list of SFS.Annotation into SFIO.Annotation"""

    if not annotations:
        return None
    ret: List[Annotation] = []
    for annotation in annotations:
        if isinstance(annotation, SFS.AttributesAnnotation):
            ret.append(
                AttributesAnnotation(
                    id=annotation.id,
                    parent_id=annotation.parent_id,
                    attributes=from_attribute_paths_sfs(annotation.attributes),
                )
            )
        elif isinstance(annotation, SFS.Box2DAnnotation):
            ret.append(
                Box2DAnnotation(
                    id=annotation.id,
                    sensor_id=annotation.sensor_id,
                    path=Box2DPath(
                        timestamps=annotation.path.timestamps,
                        values=annotation.path.values,
                    ),
                    parent_id=annotation.parent_id,
                    stationary=annotation.stationary,
                    label=annotation.label,
                    attributes=from_attribute_paths_sfs(annotation.attributes),
                )
            )
        elif isinstance(annotation, SFS.CuboidAnnotation):
            ret.append(
                CuboidAnnotation(
                    id=annotation.id,
                    path=CuboidPath(
                        # from dx, dy, dz, x, y, z, roll, pitch, yaw
                        # to   x, y, z, yaw, pitch, roll, dx, dy, dz
                        data=[
                            row[3:6] + row[8:5:-1] + row[0:3]
                            for row in annotation.path.values
                        ],
                        index=annotation.path.timestamps,
                    ),
                    label=annotation.label,
                    stationary=annotation.stationary,
                    attributes=from_attribute_paths_sfs(annotation.attributes),
                    activations=(
                        [
                            CuboidActivationPath(
                                sensor_id=activation.sensor_id,
                                timestamps=activation.timestamps,
                                durations=activation.durations,
                                cuboids=(
                                    # from dx, dy, dz, x, y, z, roll, pitch, yaw
                                    # to   x, y, z, yaw, pitch, roll, dx, dy, dz
                                    [
                                        row[3:6] + row[8:5:-1] + row[0:3]
                                        for row in activation.cuboids
                                    ]
                                    if activation.cuboids
                                    else None
                                ),
                            )
                            for activation in annotation.activations
                        ]
                        if annotation.activations
                        else None
                    ),
                    projections=(
                        [
                            CuboidProjectionPath(
                                sensor_id=projection.sensor_id,
                                timestamps=projection.timestamps,
                                boxes=projection.boxes,
                                cuboids=projection.cuboids,
                            )
                            for projection in annotation.projections
                        ]
                        if annotation.projections
                        else None
                    ),
                    parent_id=annotation.parent_id,
                )
            )
        elif isinstance(annotation, SFS.EventAnnotation):
            ret.append(
                EventAnnotation(
                    id=annotation.id,
                    start=annotation.start,
                    parent_id=annotation.parent_id,
                    label=annotation.label,
                    attributes=from_attribute_paths_sfs(annotation.attributes),
                    duration=annotation.duration,
                    sensor_id=annotation.sensor_id,
                )
            )
        elif isinstance(annotation, SFS.LabeledPointsAnnotation):
            ret.append(
                LabeledPointsAnnotation(
                    id=annotation.id,
                    label=annotation.label,
                    labeled_points=[
                        LabeledPoint(
                            sensor_id=lp.sensor_id,
                            point_ids=lp.point_ids,
                            sensor_frame=lp.sensor_frame,
                        )
                        for lp in annotation.labeled_points
                    ],
                    is_instance=annotation.is_instance,
                    parent_id=annotation.parent_id,
                )
            )
        elif isinstance(annotation, SFS.LocalizationAdjustmentAnnotation):
            ret.append(
                LocalizationAdjustmentAnnotation(
                    id=annotation.id,
                    poses=from_pose_sfs(annotation.poses),
                    parent_id=annotation.parent_id,
                )
            )
        elif isinstance(annotation, SFS.ObjectAnnotation):
            ret.append(
                ObjectAnnotation(
                    id=annotation.id,
                    parent_id=annotation.parent_id,
                    label=annotation.label,
                    attributes=from_attribute_paths_sfs(annotation.attributes),
                )
            )
        elif isinstance(annotation, SFS.Polygon2DAnnotation):
            ret.append(
                PolygonAnnotation(
                    id=annotation.id,
                    sensor_id=annotation.sensor_id,
                    path=PolygonPath(
                        timestamps=annotation.path.timestamps,
                        values=annotation.path.values,
                    ),
                    parent_id=annotation.parent_id,
                    stationary=annotation.stationary,
                    label=annotation.label,
                    attributes=from_attribute_paths_sfs(annotation.attributes),
                )
            )
        else:
            raise TypeError(f"Annotation type {type(annotation)} is not implemented")
    return ret


def to_annotations_sfs(
    annotations: Optional[List[Annotation]],
) -> Optional[List[SFS.Annotation]]:
    """Convert a list of SFIO.Annotation into SFS.Annotation"""

    if not annotations:
        return None
    ret: List[SFS.Annotation] = []
    for annotation in annotations:
        if isinstance(annotation, AttributesAnnotation):
            ret.append(
                SFS.AttributesAnnotation(
                    id=annotation.id,
                    parent_id=annotation.parent_id,
                    attributes=to_attribute_paths_sfs(annotation.attributes),
                )
            )
        elif isinstance(annotation, Box2DAnnotation):
            ret.append(
                SFS.Box2DAnnotation(
                    id=annotation.id,
                    sensor_id=annotation.sensor_id,
                    path=SFS.AnnotationPath(
                        timestamps=annotation.path.timestamps,
                        values=annotation.path.values,
                    ),
                    parent_id=annotation.parent_id,
                    stationary=annotation.stationary,
                    label=annotation.label,
                    attributes=to_attribute_paths_sfs(annotation.attributes),
                )
            )
        elif isinstance(annotation, CuboidAnnotation):
            ret.append(
                SFS.CuboidAnnotation(
                    id=annotation.id,
                    path=SFS.CuboidPath(
                        timestamps=annotation.path.index.tolist(),
                        values=annotation.path[
                            ["dx", "dy", "dz", "x", "y", "z", "roll", "pitch", "yaw"]
                        ].values.tolist(),
                    ),
                    label=annotation.label,
                    stationary=annotation.stationary,
                    attributes=to_attribute_paths_sfs(annotation.attributes),
                    activations=(
                        [
                            SFS.CuboidActivationPath(
                                sensor_id=activation.sensor_id,
                                timestamps=activation.timestamps,
                                durations=activation.durations,
                                cuboids=(
                                    # from x, y, z, yaw, pitch, roll, dx, dy, dz
                                    # to   dx, dy, dz, x, y, z, roll, pitch, yaw
                                    [
                                        row[6:9] + row[0:3] + row[5:2:-1]
                                        for row in activation.cuboids
                                    ]
                                    if activation.cuboids
                                    else None
                                ),
                            )
                            for activation in annotation.activations
                        ]
                        if annotation.activations
                        else None
                    ),
                    projections=(
                        [
                            SFS.CuboidProjectionPath(
                                sensor_id=projection.sensor_id,
                                timestamps=projection.timestamps,
                                boxes=projection.boxes,
                                cuboids=projection.cuboids,
                            )
                            for projection in annotation.projections
                        ]
                        if annotation.projections
                        else None
                    ),
                    parent_id=annotation.parent_id,
                )
            )
        elif isinstance(annotation, EventAnnotation):
            ret.append(
                SFS.EventAnnotation(
                    id=annotation.id,
                    start=annotation.start,
                    parent_id=annotation.parent_id,
                    label=annotation.label,
                    attributes=to_attribute_paths_sfs(annotation.attributes),
                    duration=annotation.duration,
                    sensor_id=annotation.sensor_id,
                )
            )
        elif isinstance(annotation, LabeledPointsAnnotation):
            ret.append(
                SFS.LabeledPointsAnnotation(
                    id=annotation.id,
                    label=annotation.label,
                    labeled_points=[
                        SFS.LabeledPoint(
                            sensor_id=lp.sensor_id,
                            point_ids=lp.point_ids,
                            sensor_frame=lp.sensor_frame,
                        )
                        for lp in annotation.labeled_points
                    ],
                    is_instance=annotation.is_instance,
                    parent_id=annotation.parent_id,
                )
            )
        elif isinstance(annotation, LocalizationAdjustmentAnnotation):
            ret.append(
                SFS.LocalizationAdjustmentAnnotation(
                    id=annotation.id,
                    poses=to_pose_sfs(annotation.poses),
                    parent_id=annotation.parent_id,
                )
            )
        elif isinstance(annotation, ObjectAnnotation):
            ret.append(
                SFS.ObjectAnnotation(
                    id=annotation.id,
                    parent_id=annotation.parent_id,
                    label=annotation.label,
                    attributes=to_attribute_paths_sfs(annotation.attributes),
                )
            )
        elif isinstance(annotation, PolygonAnnotation):
            ret.append(
                SFS.Polygon2DAnnotation(
                    id=annotation.id,
                    sensor_id=annotation.sensor_id,
                    path=SFS.AnnotationPath(
                        timestamps=annotation.path.timestamps,
                        values=annotation.path.values,
                    ),
                    parent_id=annotation.parent_id,
                    stationary=annotation.stationary,
                    label=annotation.label,
                    attributes=to_attribute_paths_sfs(annotation.attributes),
                )
            )
        else:
            raise TypeError(f"Annotation type {type(annotation)} is not implemented")
    return ret


def from_scene_spec_sfs(scene: SFS.Scene) -> Scene:
    """Convert SFS.Scene into SFIO.Scene"""

    return Scene(
        sensors=from_sensors_sfs(scene.sensors),
        annotations=from_annotations_sfs(scene.annotations),
        attributes=from_attribute_paths_sfs(scene.attributes),
        time_offset=scene.time_offset,
        time_unit=scene.time_unit,
        metadata=scene.metadata,
    )


def to_scene_spec_sfs(scene: Scene) -> SFS.Scene:
    """Convert SFIO.Scene into SFS.Scene"""

    return SFS.Scene(
        sensors=to_sensors_sfs(scene.sensors),
        annotations=to_annotations_sfs(scene.annotations),
        attributes=to_attribute_paths_sfs(scene.attributes),
        time_offset=scene.time_offset,
        time_unit=scene.time_unit,
        metadata=scene.metadata,
    )



"""Functions for working with Transform objects."""

from typing import Union
import numpy as np
import numpy.typing as npt
from scipy.spatial.transform import Rotation, Slerp

from scale_sensor_fusion_io import ITransformablePath, Pose


def interpolate_pose(a: Pose, b: Pose, factor: float) -> Pose:
    """Interpolate a Pose in between two other Poses.

    Parameters
    ----------
    a : Pose
        First Pose.
    b : Pose
        Second Pose.
    factor : float
        How close to b to interpolate in the range [0, 1]. Corresponds to the range [a, b].

    Returns
    -------
    Pose
        A Pose in between a and b.
    """
    assert 0.0 <= factor <= 1.0, f"factor {factor} is not in the range [0, 1]"

    slerp = Slerp(
        [0.0, 1.0],
        Rotation.concatenate([a.heading.as_rotation(), b.heading.as_rotation()]),
    )

    a_translation = np.asarray(a.position)
    b_translation = np.asarray(b.position)

    return Pose.from_rt(
        rotation=slerp([factor])[0],
        translation=a_translation + factor * (b_translation - a_translation),
    )


def transform_path(
    transform: Union[npt.NDArray, Pose], path: ITransformablePath
) -> ITransformablePath:
    """Apply a transformation to all poses in a ITransformablePath and return a new ITransformablePath.

    Parameters
    ----------
    transform : Union[npt.NDArray, Pose]
        Rotation and translation matrix.
    path : ITransformablePath
        Path to be transformed.

    Returns
    -------
    ITransformablePath
        Resulting path from transforming this object.
    """
    return type(path).from_matrix(transform @ path.as_matrix(), index=path.index)


def transform_points(
    transform: Union[npt.NDArray, Pose], points: npt.ArrayLike
) -> npt.NDArray[np.float64]:
    """Apply a rigid transformation to a list of points. Points should be (x, y, z).

    Parameters
    ----------
    transform : Union[npt.NDArray, Pose]
        Rotation and translation. Either a rigid transformation matrix or a Pose representing one.
    points : npt.ArrayLike
        Point cloud in the form [[x, y, z], ...].

    Returns
    -------
    npt.NDArray
        Transformed point cloud in the form [[x, y, z], ...].
    """
    if isinstance(transform, Pose):
        transform = transform.as_transform()
    return np.asarray(
        np.asarray(points, dtype=np.float64).reshape((-1, 3)) @ transform[:3, :3].T
        + transform[:3, 3]
    )


def transform_points_ignore_extra_data(
    transform: Union[npt.NDArray, Pose], points: npt.ArrayLike
) -> npt.NDArray[np.float64]:
    """Apply a rigid transformation to a list of points. Points should be at least length 3 in the
    form (x, y, z, ...). Anything past the first 3 elements is ignored (e.g. intensity, color).

    Parameters
    ----------
    transform : Union[npt.NDArray, Pose]
        Rotation and translation. Either a rigid transformation matrix or a Pose representing one.
    points : npt.ArrayLike
        Point cloud in the form [[x, y, z, ...], ...].

    Returns
    -------
    npt.NDArray
        Transformed point cloud in the form [[x, y, z, ...], ...].
    """
    points = np.asarray(points, dtype=np.float64)
    shape = points.shape if hasattr(points, 'shape') else None
    if not shape or len(shape) != 2 or shape[1] < 3:
        raise ValueError(f"invalid points shape {getattr(points, 'shape', None)}")

    if isinstance(transform, Pose):
        transform = transform.as_transform()

    points_4d = np.hstack([points[:, :3], np.ones((shape[0], 1))])
    transformed_4d = points_4d.dot(transform.T)
    return np.hstack([transformed_4d[:, :3], points[:, 3:]])



from dataclasses import dataclass
from typing import List, Literal, Optional, Union

from ..common import AnnotationID, AnnotationKind, SensorID
from ..paths import AttributePath


@dataclass
class Box2DPath:
    timestamps: List[int]
    values: List[List[float]]  # left, top, width, height


@dataclass
class Box2DAnnotation:
    id: AnnotationID
    sensor_id: SensorID
    path: Box2DPath
    type: Literal[AnnotationKind.Box2D] = AnnotationKind.Box2D
    parent_id: Optional[AnnotationID] = None
    stationary: Optional[bool] = False
    label: Optional[str] = None
    attributes: Optional[List[AttributePath]] = None



from dataclasses import asdict, dataclass
from typing import List, Tuple, Union, cast

import numpy as np
import numpy.typing as npt

from scale_sensor_fusion_io.models import (
    CameraDistortion,
    CameraIntrinsics,
    CameraSensor,
    LidarSensor,
    LidarSensorFrame,
    LidarSensorPoints,
    PosePath,
    RadarSensor,
    RadarSensorFrame,
    RadarSensorPoints,
    OdometrySensor,
    PointsSensor,
    Points,
    Scene,
    Sensor,
    SensorKind,
)

from .error import (
    DataValidationError,
    ErrorDetails,
    PathField,
    ValidationResult,
)
from .helpers import is_strictly_increasing

MICRO_IN_SEC = 1e6
MAX_FPS = 100
MIN_FPS = 1


def _handle_result(
    res: ValidationResult, error_details: List[ErrorDetails], path: List[PathField] = []
) -> None:
    if res:
        error_details.extend(
            res.details if not path else res.prepend_path(path).details
        )

    return None


def validate_pose_path(pose_path: PosePath) -> ValidationResult:
    error_details: List[ErrorDetails] = []

    pose_timestamps: List[int] = pose_path.index.tolist()
    _handle_result(
        validate_timestamps(pose_timestamps), error_details, path=["timestamps"]
    )

    if len(error_details) > 0:
        return DataValidationError(details=error_details)

    return None


def validate_timestamps(timestamps: List[int]) -> ValidationResult:
    error_details: List[ErrorDetails] = []
    if any(ts < 0 for ts in timestamps):
        error_details.append(
            ErrorDetails.from_msg(
                "timestamps must not be negative",
            )
        )
    if not is_strictly_increasing(timestamps):
        error_details.append(
            ErrorDetails.from_msg(
                "timestamps must be strictly increasing",
            )
        )

    max_ts_diff = MICRO_IN_SEC / MIN_FPS
    max_init_frame_ts = max_ts_diff * 100  # allow a padding of 100 frames
    if timestamps[0] > max_init_frame_ts:
        error_details.append(
            ErrorDetails.from_msg(
                f"timestamps must be normalized: {timestamps[0]} > {max_init_frame_ts}",
            )
        )

    if len(error_details) > 0:
        return DataValidationError(details=error_details)
    return None


def validate_fps(timestamps: List[int], max_fps: int = MAX_FPS) -> ValidationResult:
    error_details: List[ErrorDetails] = []

    # compute approximate fps from timestamps
    if len(timestamps) < 2:
        return None

    avg_ts_diffs = cast(
        float, np.mean([t2 - t1 for t1, t2 in zip(timestamps, timestamps[1:])])
    )
    fps = float(MICRO_IN_SEC) / avg_ts_diffs if avg_ts_diffs != 0 else np.inf

    if fps > max_fps:
        error_details.append(
            ErrorDetails.from_msg(
                f"approximate fps is too high: {fps} > {max_fps}",
            )
        )

    if len(error_details) > 0:
        return DataValidationError(details=error_details)

    return None


def validate_points_positions(positions: npt.NDArray[np.float32]) -> ValidationResult:
    error_details: List[ErrorDetails] = []

    if positions.dtype != np.float32:
        error_details.append(
            ErrorDetails.from_msg(
                f"Positions must have dtype float32: is {positions.dtype}",
            )
        )

    shape = positions.shape if hasattr(positions, 'shape') else None
    if not shape or len(shape) != 2 or shape[1] != 3:
        error_details.append(
            ErrorDetails.from_msg(
                f"Positions must have shape (n, 3): is ({positions.shape})",
            )
        )

    if len(error_details) > 0:
        return DataValidationError(details=error_details)

    return None


def validate_points_colors(
    colors: npt.NDArray[np.uint8], expected_length: int
) -> ValidationResult:
    error_details: List[ErrorDetails] = []

    if colors.dtype != np.uint8:
        error_details.append(
            ErrorDetails.from_msg(
                f"Colors must have dtype uint8: is {colors.dtype}",
            )
        )

    shape = colors.shape if hasattr(colors, 'shape') else None
    if not shape or len(shape) != 2 or shape[1] != 3:
        error_details.append(
            ErrorDetails.from_msg(
                f"Colors must have shape (n, 3): is ({colors.shape})",
            )
        )

    if len(colors) != expected_length:
        error_details.append(
            ErrorDetails.from_msg(
                f"length of colors ({len(colors)}) should match length of positions: ({expected_length})",
            )
        )

    if len(error_details) > 0:
        return DataValidationError(details=error_details)

    return None


def validate_points_intensities(
    intensities: npt.NDArray[np.uint8], expected_length: int
) -> ValidationResult:
    error_details: List[ErrorDetails] = []

    if intensities.dtype != np.uint8:
        error_details.append(
            ErrorDetails.from_msg(
                f"Intensities must have dtype uint8: is {intensities.dtype}",
            )
        )

    if len(intensities.shape) != 1:
        error_details.append(
            ErrorDetails.from_msg(
                f"Intensities must have shape (n, 1): is ({intensities.shape})",
            )
        )

    if len(intensities) != expected_length:
        error_details.append(
            ErrorDetails.from_msg(
                f"length of intensities ({len(intensities)}) should match length of positions: ({expected_length})",
            )
        )

    if len(error_details) > 0:
        return DataValidationError(details=error_details)

    return None


def validate_points_timestamps(
    timestamps: Union[npt.NDArray[np.uint32], npt.NDArray[np.uint64]],
    expected_length: int,
) -> ValidationResult:
    error_details: List[ErrorDetails] = []

    if timestamps.dtype != np.uint32 and timestamps.dtype != np.uint64:
        error_details.append(
            ErrorDetails.from_msg(
                f"Timestamps must have dtype uint32 or uint64: is {timestamps.dtype}",
            )
        )

    shape: Tuple[int, ...] = tuple(timestamps.shape) 
    if not shape or len(shape) != 1:
        error_details.append(
            ErrorDetails.from_msg(
                f"Timestamps must have shape (n, 1): is ({getattr(timestamps, 'shape', None)})",
            )
        )

    ts_arr = np.asarray(timestamps)
    ts_length = len(ts_arr)
    if ts_length != expected_length:
        error_details.append(
            ErrorDetails.from_msg(
                f"length of timestamps ({ts_length}) should match length of positions: ({expected_length})",
            )
        )

    if len(error_details) > 0:
        return DataValidationError(details=error_details)

    return None


def validate_points_directions(
    directions: npt.NDArray[np.float32], expected_length: int
) -> ValidationResult:
    error_details: List[ErrorDetails] = []

    if directions.dtype != np.float32:
        error_details.append(
            ErrorDetails.from_msg(
                f"Directions must have dtype float32: is {directions.dtype}",
            )
        )

    shape = directions.shape if hasattr(directions, 'shape') else None
    if len(shape) != 2 or shape[1] != 3:
        error_details.append(
            ErrorDetails.from_msg(
                f"Directions must have shape (n, 3): is ({directions.shape})",
            )
        )

    if len(directions) != expected_length:
        error_details.append(
            ErrorDetails.from_msg(
                f"length of directions ({len(directions)}) should match length of positions: ({expected_length})",
            )
        )

    if len(error_details) > 0:
        return DataValidationError(details=error_details)

    return None


def validate_points_lengths(
    lengths: npt.NDArray[np.float32], expected_length: int
) -> ValidationResult:
    error_details: List[ErrorDetails] = []

    if lengths.dtype != np.float32:
        error_details.append(
            ErrorDetails.from_msg(
                f"Lengths must have dtype float32: is {lengths.dtype}",
            )
        )

    if len(lengths.shape) != 1:
        error_details.append(
            ErrorDetails.from_msg(
                f"Lengths must have shape (n, 1): is ({lengths.shape})",
            )
        )

    if len(lengths) != expected_length:
        error_details.append(
            ErrorDetails.from_msg(
                f"length of lengths ({len(lengths)}) should match length of positions: ({expected_length})",
            )
        )

    if len(error_details) > 0:
        return DataValidationError(details=error_details)

    return None


def validate_lidar(sensor: LidarSensor) -> ValidationResult:
    """Validate lidar sensor"""
    error_details: List[ErrorDetails] = []
    content_timestamps = [frame.timestamp for frame in sensor.frames]

    _handle_result(
        validate_timestamps(content_timestamps), error_details, path=["frames"]
    )
    _handle_result(validate_fps(content_timestamps), error_details, path=["frames"])

    # pose validation
    _handle_result(
        validate_pose_path(sensor.poses),
        error_details,
        path=["poses"],
    )

    for frame_num, frame in enumerate(sensor.frames):
        _handle_result(
            validate_points_positions(frame.points.positions),
            error_details,
            path=["frames", frame_num, "points", "positions"],
        )

        pos_length = len(frame.points.positions)
        if frame.points.colors is not None:
            _handle_result(
                validate_points_colors(frame.points.colors, expected_length=pos_length),
                error_details,
                path=["frames", frame_num, "points", "colors"],
            )

        if frame.points.intensities is not None:
            _handle_result(
                validate_points_intensities(
                    frame.points.intensities, expected_length=pos_length
                ),
                error_details,
                path=["frames", frame_num, "points", "intensities"],
            )

        if frame.points.timestamps is not None:
            _handle_result(
                validate_points_timestamps(
                    frame.points.timestamps, expected_length=pos_length
                ),
                error_details,
                path=["frames", frame_num, "points", "timestamps"],
            )

            next_frame_ts = (
                sensor.frames[frame_num + 1].timestamp
                if frame_num < len(sensor.frames) - 1
                else None
            )

            ts_array = np.asarray(frame.points.timestamps)
            min_points_ts = np.min(ts_array) if len(ts_array) > 0 else 0

            # NOTE: this is the correct validation, but it is too strict for now. We can add this back as a warning once we support warnings vs errors
            # if (
            #     frame.timestamp > min_points_ts
            #     or next_frame_ts
            #     and max_points_ts > next_frame_ts
            # ):
            #     error_details.append(
            #         ErrorDetails.from_msg(
            #             f"point timestamps (range {min_points_ts} -> {max_points_ts}) must be included within consecutive frame timestamps (range {frame.timestamp} -> {next_frame_ts})",
            #             path=["frames", frame_num, "points"],
            #         )
            #     )

            # Simpler check: frame timestamp must be less the min point timestamp
            if frame.timestamp > min_points_ts:
                error_details.append(
                    ErrorDetails.from_msg(
                        f"frame timestamp ({frame.timestamp}) must be less than min point timestamp ({min_points_ts})",
                        path=["frames", frame_num, "points", "timestamps"],
                    )
                )

    if len(error_details) > 0:
        return DataValidationError(details=error_details)

    return None


def validate_radar(sensor: RadarSensor) -> ValidationResult:
    """Validate radar sensor"""
    error_details: List[ErrorDetails] = []
    content_timestamps = [frame.timestamp for frame in sensor.frames]

    _handle_result(
        validate_timestamps(content_timestamps), error_details, path=["frames"]
    )
    _handle_result(validate_fps(content_timestamps), error_details, path=["frames"])

    # pose validation
    _handle_result(
        validate_pose_path(sensor.poses),
        error_details,
        path=["poses"],
    )

    for frame_num, frame in enumerate(sensor.frames):
        _handle_result(
            validate_points_positions(frame.points.positions),
            error_details,
            path=["frames", frame_num, "points", "positions"],
        )

        pos_length = len(frame.points.positions)
        if frame.points.directions is not None:
            _handle_result(
                validate_points_directions(
                    frame.points.directions, expected_length=pos_length
                ),
                error_details,
                path=["frames", frame_num, "points", "directions"],
            )

        if frame.points.lengths is not None:
            _handle_result(
                validate_points_lengths(
                    frame.points.lengths, expected_length=pos_length
                ),
                error_details,
                path=["frames", frame_num, "points", "lengths"],
            )

        if frame.points.timestamps is not None:
            _handle_result(
                validate_points_timestamps(
                    frame.points.timestamps, expected_length=pos_length
                ),
                error_details,
                path=["frames", frame_num, "points", "timestamps"],
            )

            next_frame_ts = (
                sensor.frames[frame_num + 1].timestamp
                if frame_num < len(sensor.frames) - 1
                else None
            )

            ts_array = np.asarray(frame.points.timestamps)
            min_points_ts = np.min(ts_array) if len(ts_array) > 0 else 0

            # NOTE: this is the correct validation, but it is too strict for now. We can add this back as a warning once we support warnings vs errors
            # if (
            #     frame.timestamp > min_points_ts
            #     or next_frame_ts
            #     and max_points_ts > next_frame_ts
            # ):
            #     error_details.append(
            #         ErrorDetails.from_msg(
            #             f"point timestamps (range {min_points_ts} -> {max_points_ts}) must be included within consecutive frame timestamps (range {frame.timestamp} -> {next_frame_ts})",
            #             path=["frames", frame_num, "points"],
            #         )
            #     )

            # Simpler check: frame timestamp must be less the min point timestamp
            if frame.timestamp > min_points_ts:
                error_details.append(
                    ErrorDetails.from_msg(
                        f"frame timestamp ({frame.timestamp}) must be less than min point timestamp ({min_points_ts})",
                        path=["frames", frame_num, "points", "timestamps"],
                    )
                )

    if len(error_details) > 0:
        return DataValidationError(details=error_details)

    return None


def validate_camera(sensor: CameraSensor) -> ValidationResult:
    """Validate camera sensor"""
    error_details: List[ErrorDetails] = []

    # camera content
    content_timestamps: List[int] = []
    content_timestamps_path: List[PathField] = []
    if sensor.video:
        content_timestamps = sensor.video.timestamps
        content_timestamps_path = ["video"]
    elif sensor.images:
        content_timestamps = [img.timestamp for img in sensor.images]
        content_timestamps_path = ["images"]
    else:
        error_details.append(
            ErrorDetails.from_msg('Exactly one of "images" or "video" expected')
        )

    _handle_result(
        validate_timestamps(content_timestamps),
        error_details,
        path=content_timestamps_path,
    )
    _handle_result(
        validate_fps(content_timestamps),
        error_details,
        path=content_timestamps_path,
    )

    # pose validation
    _handle_result(validate_pose_path(sensor.poses), error_details, path=["poses"])

    if len(error_details) > 0:
        return DataValidationError(details=error_details)

    return None


def validate_points_sensor(sensor: PointsSensor) -> ValidationResult:
    # pose validation
    error_details: List[ErrorDetails] = []

    _handle_result(
        validate_points_positions(sensor.points.positions),
        error_details,
        path=["points", "positions"],
    )

    if sensor.points.colors is not None:
        _handle_result(
            validate_points_colors(
                sensor.points.colors, expected_length=len(sensor.points.positions)
            ),
            error_details,
            path=["points", "colors"],
        )

    if len(error_details) > 0:
        return DataValidationError(details=error_details)

    return None


def validate_odometer(sensor: OdometrySensor) -> ValidationResult:
    # pose validation
    error_details: List[ErrorDetails] = []

    _handle_result(validate_pose_path(sensor.poses), error_details, path=["poses"])

    if len(error_details) > 0:
        return DataValidationError(details=error_details)

    return None


def validate_sensor(sensor: Sensor) -> ValidationResult:
    error_details: List[ErrorDetails] = []
    if sensor.type is SensorKind.Camera:
        _handle_result(validate_camera(sensor), error_details)
    elif sensor.type is SensorKind.Lidar:
        _handle_result(validate_lidar(sensor), error_details)
    elif sensor.type is SensorKind.Radar:
        _handle_result(validate_radar(sensor), error_details)
    elif sensor.type is SensorKind.Points:
        _handle_result(validate_points_sensor(sensor), error_details)
    elif sensor.type is SensorKind.Odometer:
        _handle_result(validate_odometer(sensor), error_details)
    else:
        error_details.append(
            ErrorDetails(
                path=["type"], errors=[f"Invalid sensor type provided: {sensor.type}"]
            )
        )
    if len(error_details) > 0:
        return DataValidationError(details=error_details)

    return None


def validate_scene(scene: Scene) -> ValidationResult:
    """Validate scene"""
    error_details: List[ErrorDetails] = []
    if scene.sensors:
        for sensor in scene.sensors:
            _handle_result(
                validate_sensor(sensor),
                error_details,
                path=["sensors", str(sensor.id)],
            )

        if len(scene.sensors) != len(set([sensor.id for sensor in scene.sensors])):
            error_details.append(
                ErrorDetails.from_msg("Sensor ids must be unique", path=["sensors"])
            )

    if scene.time_unit != "microseconds":
        error_details.append(
            ErrorDetails.from_msg(
                f"Invalid time unit provided: {scene.time_unit}. Expected: microseconds"
            )
        )

    if scene.time_offset is not None and scene.time_offset < 0:
        error_details.append(
            ErrorDetails.from_msg("Scene timestamp must not be negative")
        )

    if len(error_details) > 0:
        return DataValidationError(details=error_details)
    return None



from dataclasses import dataclass
from typing import List, Literal, Optional

from scale_sensor_fusion_io.models.paths.cuboid_path import CuboidPath

from ..common import AnnotationID, AnnotationKind
from ..paths import PosePath
"""
The LocalizationAdjustmentAnnotation represents a PosePath applied a scene to fix localization issues or convert from ego to world coordinates.
"""


@dataclass
class LayerMetadata:
    layer_type: Literal["base", "cuboid"]
    order: int
    name: str
    cuboids: Optional[List[CuboidPath]] = None
    algorithm: Optional[
        Literal["position", "position-rotation", "z-level", "icp"]
    ] = None

@dataclass
class LocalizationAdjustmentAnnotation:
    id: AnnotationID
    poses: PosePath
    type: Literal[AnnotationKind.LocalizationAdjustment] = AnnotationKind.LocalizationAdjustment
    parent_id: Optional[AnnotationID] = None
    layer_metadata: Optional[LayerMetadata] = None



from dataclasses import dataclass
from typing import List, Optional, Union

from ..common import SensorID


@dataclass
class AttributePath:
    name: str
    timestamps: List[int]
    values: List[Union[str, int, List[str]]]
    sensor_id: Optional[SensorID] = None
    static: bool = False



from dataclasses import dataclass
from typing import List, Literal, Optional, Union

import numpy as np
import numpy.typing as npt

from ..common import AnnotationID, AnnotationKind, SensorID


# Define LabeledPointsAnnotationLabeledPoint dataclass
@dataclass
class LabeledPoint:
    sensor_id: SensorID
    point_ids: npt.NDArray[np.uint32]
    sensor_frame: Optional[int] = None


# Define LabeledPointsAnnotation dataclass
@dataclass
class LabeledPointsAnnotation:
    id: AnnotationID
    label: str
    labeled_points: List[LabeledPoint]
    is_instance: bool = False
    type: Union[Literal[AnnotationKind.LabeledPoints],Literal[AnnotationKind.PointSegmentation]] = AnnotationKind.PointSegmentation
    parent_id: Optional[AnnotationID] = None



from typing import Union

from .points import *
from .camera import *
from .lidar import *
from .radar import *
from .odometry import *

Sensor = Union[
    PointsSensor,
    CameraSensor,
    LidarSensor,
    RadarSensor,
    OdometrySensor,
]
